//network.cpp
//basic game hosting/joining and message sending code based on Network_Tool

#include <network_tool.h>
#include <stdio.h>
#include <string.h>
#include "gametime.h"
#include "config.h"
#include "network.h"
#include "error.h"
#include "gameframe.h"
#include "controls.h"
#include "gamesound.h"
#include "carphysics.h"
#include "interfaceutil.h"
#include "gamemem.h"
#include "Reggie.h"
#include "gamesystem.h"
#include "initexit.h"

extern ReggieRef gReggieRef;
#define kGameName "Redline"
#define kConnectionTimeOut 5.0

#define S32ChangeToNative(value) ((value)=EndianS32_BtoN(value))
#define U32ChangeToNative(value) ((value)=EndianU32_BtoN(value))
#define S32ChangeToNet(value) ((value)=EndianS32_NtoB(value))
#define U32ChangeToNet(value) ((value)=EndianU32_NtoB(value))

NT_SessionRef gSessionRef=nil;

typedef struct{
	QStub qstub;
	tNetworkPacket packet;
}tQueuedPacket;

Queue gPacketQueue;
int gInternetAvailable=true;
int gTerminateSession=false;

void NetworkLockOutNewPlayers()
{
//	HandleError(NT_SetSessionLock(gSessionRef,true));
	int err=NT_SetSessionLock(gSessionRef,true);
	PrintConsoleString("Locking session. returns %d",err);
}

void NetworkUnlockOutNewPlayers()
{
//	HandleError(NT_SetSessionLock(gSessionRef,false));
	int err=NT_SetSessionLock(gSessionRef,false);
	PrintConsoleString("Unlocking session. returns %d",err);
}

void NetworkClearPacketQueue()
{
	int temp=gTerminateSession;
	gTerminateSession=false;
	tNetworkPacket packet;
	while(NetworkReceivePacket(&packet))
		if(packet.data)
			MemoryFreeBlock(packet.data);
	gTerminateSession=temp;
}

void NetworkGetStatusString(char *str)
{
	NT_ProblemInfo prob;
	if(NT_GetProblemInfo(gSessionRef,0,&prob))
	{
		if(prob.timeout<10)
			str[0]='.';
		else if(prob.timeout<50)
			str[0]='t';
		else
			str[0]='T';
		if(prob.latency<10)
			str[1]='.';
		else if(prob.latency<50)
			str[1]='l';
		else
			str[1]='L';
		if(prob.stalled<10)
			str[2]='.';
		else if(prob.stalled<50)
			str[2]='s';
		else
			str[2]='S';
		if(prob.backlog<10)
			str[3]='.';
		else if(prob.backlog<50)
			str[3]='b';
		else
			str[3]='B';
		if(prob.failure<10)
			str[4]='.';
		else if(prob.failure<50)
			str[4]='f';
		else
			str[4]='F';
		if(prob.errors<10)
			str[5]='.';
		else if(prob.errors<50)
			str[5]='e';
		else
			str[5]='E';
		str[6]='\0';
	}
}

void NetworkQueuePacket(tNetworkPacket *insert)
{
	tQueuedPacket *q=(tQueuedPacket*)MemoryAllocateZeroedBlock(sizeof(tQueuedPacket));
	q->packet=*insert;
	QueueInsert(&gPacketQueue,(QStubPtr)q);
}

//#define PACKET_DUMP

#ifdef PACKET_DUMP
void DumpPacket(tNetworkPacket *packet)
{
	char st[256];
	sprintf(st,"what: %d",packet->what);
	PrintConsoleString(st);
	sprintf(st,"from: %d",packet->from);
	PrintConsoleString(st);
	sprintf(st,"size: 0x%x",packet->size);
	PrintConsoleString(st);
	
	int line=0;
	
	while(line<=packet->size/16)
	{
		char dataStr[256];
								
		sprintf(dataStr,"%04x  ",line*16);

		for(int pos=line*16;pos<line*16+16;pos++)
			if(pos<packet->size)
				sprintf(dataStr,"%s%02x",dataStr,*((UInt8*)packet->data+pos));
			else
				sprintf(dataStr,"%s  ",dataStr);

		sprintf(dataStr,"%s  ",dataStr);
				
		for(int pos=line*16;pos<line*16+16;pos++)
			if(pos<packet->size)
				if(*((UInt8*)packet->data+pos)>=32)
					sprintf(dataStr,"%s%c",dataStr,*((UInt8*)packet->data+pos));
				else
					sprintf(dataStr,"%s.",dataStr);
			else
				sprintf(dataStr,"%s ",dataStr);

		PrintConsoleString(dataStr);
		line++;
	}
	
	PrintConsoleString("");
}
#endif

void NTEventHandler(NT_SessionRef sessionRef,void *refcon,NT_SessionEventType event,NT_MemberIDType member)
{
	switch(event)
	{
		case eSessionEventSessionStart:{
				PrintConsoleString("eSessionEventSessionStart");				
				tNetworkPacket packet;
				packet.what=kMessageTypePlayerJoined;
				packet.from=0;
				packet.size=0;
				packet.data=NULL;	
				NetworkQueuePacket(&packet);
			}
			break;
		case eSessionEventMemberJoin:{
				PrintConsoleString("eSessionEventMemberJoin, member=%d",member);				
				tNetworkPacket packet;
				packet.what=kMessageTypePlayerJoined;
				packet.from=member;
				packet.size=0;
				packet.data=NULL;	
				NetworkQueuePacket(&packet);
				NT_ToleranceInfo tolerance;
				NT_GetToleranceInfo(gSessionRef,&tolerance);
				PrintConsoleString("Tolerance info:");
				PrintConsoleString("avgPackets %d.",tolerance.avgPackets);
				PrintConsoleString("avgRequests %d.",tolerance.avgRequests);
				PrintConsoleString("maxRequests %d.",tolerance.maxRequests);
				PrintConsoleString("minHeartbeat %d.",tolerance.minHeartbeat);
				PrintConsoleString("avgHeartbeat %d.",tolerance.avgHeartbeat);
				PrintConsoleString("maxHeartbeat %d.",tolerance.maxHeartbeat);
				PrintConsoleString("maxTimeout %d.",tolerance.maxTimeout);
				PrintConsoleString("maxFailures %d.",tolerance.maxFailures);
				PrintConsoleString("maxErrors %d.",tolerance.maxErrors);
				PrintConsoleString("minLinger %d.",tolerance.minLinger);
			}
			break;
		case eSessionEventMemberLeave:{
				PrintConsoleString("eSessionEventMemberLeave, member=%d",member);
				tNetworkPacket packet;
				if(member==0)
					packet.what=kMessageTypeGameTerminated;
				else
					packet.what=kMessageTypePlayerLeft;	
				packet.from=member;
				packet.size=0;
				packet.data=NULL;					
				NetworkQueuePacket(&packet);
				NT_LatencyInfo latency;
				if(NT_GetLatencyInfo(gSessionRef,member,&latency))
				{
					PrintConsoleString("Latency info for dropped member:");
					PrintConsoleString("minLinkRTT %d.",latency.minLinkRTT);
					PrintConsoleString("avgLinkRTT %d.",latency.avgLinkRTT);
					PrintConsoleString("p80LinkRTT %d.",latency.p80LinkRTT);
					PrintConsoleString("devLinkRTT %d.",latency.devLinkRTT);
					PrintConsoleString("wAvgLinkRTT %d.",latency.wAvgLinkRTT);
					PrintConsoleString("wDevLinkRTT %d.",latency.wDevLinkRTT);
					PrintConsoleString("minTranRTT %d.",latency.minTranRTT);
					PrintConsoleString("avgTranRTT %d.",latency.avgTranRTT);
					PrintConsoleString("p80TranRTT %d.",latency.p80TranRTT);
					PrintConsoleString("maxTranRTT %d.",latency.maxTranRTT);
					PrintConsoleString("devTranRTT %d.",latency.devTranRTT);
					PrintConsoleString("wAvgTranRTT %d.",latency.wAvgTranRTT);
					PrintConsoleString("wDevTranRTT %d.",latency.wDevTranRTT);
				}
				NT_ProblemInfo prob;
				if(NT_GetProblemInfo(gSessionRef,member,&prob))
				{
					PrintConsoleString("Problem info for dropped member:");
					PrintConsoleString("timeout %d.",prob.timeout);
					PrintConsoleString("latency %d.",prob.latency);
					PrintConsoleString("stalled %d.",prob.stalled);
					PrintConsoleString("backlog %d.",prob.backlog);
					PrintConsoleString("failure %d.",prob.failure);
					PrintConsoleString("errors %d.",prob.errors);
				}
			}
			break;
		case eSessionEventSessionEnd:{		
				PrintConsoleString("eSessionEventSessionEnd");
				tNetworkPacket packet;
				packet.what=kMessageTypeGameTerminated;
				packet.from=member;
				packet.size=0;
				packet.data=NULL;					
				NetworkQueuePacket(&packet);
			}
			break;
		case eSessionEventRecvPacket:{
				UInt32 data[256];
				UInt16 len;
				if(NT_RecvPacket(gSessionRef,&member,(char*)data,1024,&len))
				{
					tNetworkPacket packet;
					packet.data=MemoryAllocateBlock(len-4);
					MemoryMove(packet.data,data+1,len-4);
					packet.size=len-4;
					packet.from=member;
					packet.what=*data;
					S32ChangeToNative(packet.what);
					if(packet.what==kMessageTypeGameTerminated)
						gTerminateSession=true;
					/*#ifdef PACKET_DUMP
					PrintConsoleString("Packet Received:");
					DumpPacket(&packet);
					#endif
				*/
					NetworkQueuePacket(&packet);
				}
			}
			break;
		case eSessionEventRecvRequest:{
				UInt32 data[256];
				UInt16 len;
				if(NT_RecvRequest(gSessionRef,&member,(char*)data,1024,&len))
				{
					NT_SendResponse(gSessionRef,NULL,0);
					tNetworkPacket packet;
					packet.data=MemoryAllocateBlock(len-4);
					MemoryMove(packet.data,data+1,len-4);
					packet.size=len-4;
					packet.from=member;
					packet.what=*data;
					S32ChangeToNative(packet.what);
					if(packet.what==kMessageTypeGameTerminated)
						gTerminateSession=true;
					
					#ifdef PACKET_DUMP
					PrintConsoleString("Request Packet Received:");
					DumpPacket(&packet);
					#endif

					NetworkQueuePacket(&packet);
				}
			}
			break;
		case eSessionEventRecvResponse:{
				UInt32 data[256];
				UInt16 len;
				NT_RecvResponse(gSessionRef,&member,(char*)data,1024,&len);
			}			
			break;
	}
}

void NTErrorHandler(NT_SessionRef sessionRef,void *refcon,NT_SessionProblemType problem,NT_SessionSeverityType severity,NT_MemberIDType member,Bool8 *disconnectSelf,Bool8 *disconnectMember)
{
	PrintConsoleString("NT Error %d",problem);

	tNetworkPacket packet;
	packet.data=NULL;
	packet.size=0;
	packet.from=member;
	if(member==0)
		packet.what=kMessageTypeGameTerminated;
	else
		packet.what=kMessageTypePlayerLeft;	
	NetworkQueuePacket(&packet);
}
	
void NetworkPreSession()
{
	HandleError(NT_SessionStartParam(&gSessionRef,NTEventHandler,NULL,NULL,NULL,6,kGameName,""));	
}

float NetworkGetPlayerPing(int netID)
{
	NT_MemberInfo info;
	NT_GetSelfMemberInfo(gSessionRef,&info);
	if(netID==info.member||netID==kMessageNoRecipients)
		return 0;
		
	NT_LatencyInfo latency;
	if(NT_GetLatencyInfo(gSessionRef,netID,&latency));
	{
	/*	PrintConsoleString("Player %d latency.",netID);
		PrintConsoleString("minLinkRTT %d.",latency.minLinkRTT);
		PrintConsoleString("avgLinkRTT %d.",latency.avgLinkRTT);
		PrintConsoleString("p80LinkRTT %d.",latency.p80LinkRTT);
		PrintConsoleString("devLinkRTT %d.",latency.devLinkRTT);
		PrintConsoleString("wAvgLinkRTT %d.",latency.wAvgLinkRTT);
		PrintConsoleString("wDevLinkRTT %d.",latency.wDevLinkRTT);
		PrintConsoleString("minTranRTT %d.",latency.minTranRTT);
		PrintConsoleString("avgTranRTT %d.",latency.avgTranRTT);
		PrintConsoleString("p80TranRTT %d.",latency.p80TranRTT);
		PrintConsoleString("maxTranRTT %d.",latency.maxTranRTT);
		PrintConsoleString("devTranRTT %d.",latency.devTranRTT);
		PrintConsoleString("wAvgTranRTT %d.",latency.wAvgTranRTT);
		PrintConsoleString("wDevTranRTT %d.",latency.wDevTranRTT);*/
		return latency.avgLinkRTT*0.001;	
	}
	return 0;
}
	
//Host a new Network Game
void HostNetGame(int maxPlayers,char *pass)
{
	HandleError(NT_SessionStartParam(&gSessionRef,NTEventHandler,NULL,NULL,NULL,maxPlayers,kGameName,pass));	
	gTerminateSession=false;
}

void NetworkDisconnectPlayer(int netID)
{
	NT_MemberInfo info;
	if(NT_GetMemberInfo(gSessionRef,netID,&info))
	{
		HandleError(NT_SessionBan(gSessionRef,&info.address,false,NULL));
		NetworkSendPacket(kMessageTypeKicked,NULL,0,kMessagePriorityHigh,netID);
	}
}

void NetworkChangePassword(char *pass)
{
	HandleError(NT_SetSessionPassword(gSessionRef,pass));
}

//join a network game at the ip address given in the string address.
//returns an error code or 0. passes the player's id in id.
int JoinNetGame(char *address,int *id,int carID,char *errorString)
{
	gTerminateSession=false;
	if(!*address)
	{
		sprintf(errorString,"No Address Entered.");
		return false;
	}
	PrintConsoleString(address);
	NetworkAddress addy;
	Result32 err=SimpleResolve(eNetworkStackTCPIP,address,&addy);
	if(err)
	{
		sprintf(errorString,"Can't resolve host. (%d)",err);
		return false;
	}

	NetworkClearPacketQueue();
	HandleError(NT_SessionRejoin(gSessionRef,&addy,kGameName,NULL,&err));
	
	if(err==eNetworkToolErrorBadPass)
	{
		err=0;
		char password[256];
		if(InterfaceGetUserInputString("Please enter Game Password",password,256,true,true))
		{
			NetworkClearPacketQueue();
			HandleError(NT_SessionRejoin(gSessionRef,&addy,kGameName,password,&err));
		}
		else
			return false;
	}
	if(err)
	{
		switch(err)
		{
			case eNetworkToolErrorNoServer:
				sprintf(errorString,"No Server Running on that Machine",err);
				break;
			case eNetworkToolErrorTimeout:
				sprintf(errorString,"Network_Tool Timeout",err);
				break;
			case eNetworkToolErrorLatency:
				sprintf(errorString,"Latency is too high.",err);
				break;
			case eNetworkToolErrorBanned:
				sprintf(errorString,"You are banned from this host.",err);
				break;
			case eNetworkToolErrorTooMany:
				sprintf(errorString,"The game is full.",err);
				break;
			case eNetworkToolErrorBadPass:
				sprintf(errorString,"Wrong Password",err);
				break;
			case eNetworkToolErrorLocked:
				sprintf(errorString,"Can't join in the middle of a game.",err);
				break;
			default:
				sprintf(errorString,"Connection Error. (%d)",err);
		}
		return false;
	}
		
	float startTime=TimeGetSeconds();
	
	while(TimeGetSeconds()<startTime+kConnectionTimeOut)
	{
		tNetworkPacket message;
		while(NetworkReceivePacket(&message))
		{			
			char str[256];

			//wait for servers response
			if(message.what==kMessageTypeVersion)
			{
				UInt32 reply=kVersion;
				U32ChangeToNet(reply);
				NetworkSendPacket(kMessageTypeVersion,&reply,sizeof(UInt32),kMessagePriorityHigh,message.from);				
			}
			if(message.what==kMessageTypeID)
			{
				*id=*(UInt8*)message.data;
				tCarIDMessage reply;
				reply.playerID=*id;
				reply.color=gConfig->lastColor;
//				reply.checksum=FileGetChecksum(carID);
//				U32ChangeToNet(reply.checksum);
				strcpy(reply.carName,FileGetName(carID));
				strcpy(reply.playerName,gConfig->playerName);
				NetworkSendPacket(kMessageTypeCarID,&reply,sizeof(tCarIDMessage),kMessagePriorityHigh,message.from);
				return true;
			}
			if(message.what==kMessageTypeJoinDenied)
			{	
				if(*((UInt8*)message.data)==kJoinDeniedInProgress)
					sprintf(errorString,"Can't join in the middle of a game.",err);
				else if(*((UInt8*)message.data)==kJoinDeniedVersion)
					sprintf(errorString,"Version Mismatch.",err);
				else
					sprintf(errorString,"Host rejected connection.",err);		
				PrintConsoleString("Rejected. Exiting.");			
				return false;
			}
			NetworkDisposePacket(&message);
		}
	}
	sprintf(errorString,"Connection Timed out.",err);
	return false;
}

//exits a network game
void ExitNetGame()
{
	if(gSessionRef)
	{
		HandleError(NT_SessionLeave(gSessionRef));
		gSessionRef=nil;
	}
}

//sends the data in *message to other players on the network
void NetworkSendPacket(int messageType,void *message,int size,int priority,int to)
{
	if(messageType==kMessageTypeGameTerminated)
		gTerminateSession=true;
		
	if(to==kMessageNoRecipients)
		return;
	
	if(size>1020)
		FailWithErrorString("Network_Tool Can't send packet >1K");

	NT_MemberInfo info;
	NT_GetSelfMemberInfo(gSessionRef,&info);
	NT_MemberListType self=0x00000001<<info.member;
	
	UInt32 data[256];
	MemoryMove(data+1,message,size);
	*data=messageType;
	U32ChangeToNet(*data);
	
	NT_MemberListType members;
	switch(to)
	{
		case kMessageSendToAll:			members=eMemberListAll;			break;
		case kMessageSendToAllButSelf:	members=eMemberListAll^self;	break;
		case kMessageSendToHostOnly:	members=0x00000001;				break;
		default: members=0x00000001<<to;
	}
		
	if(members&self)
	{
		tNetworkPacket packet;
		packet.data=MemoryAllocateBlock(size);
		MemoryMove(packet.data,message,size);
		packet.size=size;
		packet.from=info.member;
		packet.what=messageType;	

		NetworkQueuePacket(&packet);	
	}
	
	Result32 err=0;
	if(priority<=kMessagePriorityNormal)
		err=NT_SendPacket(gSessionRef,members,(char*)data,size+4);
	else
	{
		NT_SessionInfo sessInfo;
		NT_MemberInfo member;
		
		NT_GetSessionInfo(gSessionRef,&sessInfo);
		for(int i=0;i<32;i++)
			if((members&(0x00000001<<i))&&(i!=info.member))
				if(NT_GetMemberInfo(gSessionRef,i,&member))
					err=NT_SendRequest(gSessionRef,i,(char*)data,size+4);
	}
	if(err)
	{
		PrintConsoleString("Error Sending Packet (%d)",err);		
		gTerminateSession=true;
	}
}

int NetworkReceivePacket(tNetworkPacket* packet)
{
	if(gTerminateSession)
	{
		packet->what=kMessageTypeGameTerminated;
		packet->data=NULL;
		return true;
	}
	else if(QStubPtr rcv=QueueRemove(&gPacketQueue))
	{
		*packet=((tQueuedPacket*)rcv)->packet;
		MemoryFreeBlock(rcv);
		return true;
	}
	else return false;
}

void NetworkDisposePacket(tNetworkPacket* packet)
{
	if(packet->data)
		MemoryFreeBlock(packet->data);
}

void NetworkInit()
{
	HandleError(NT_Open());
	QueueCreate(&gPacketQueue);
}

void NetworkIdle()
{
	NT_Idle();
}

void NetworkExit()
{
	NT_Close();
	QueueEmpty(&gPacketQueue);
}

//wait for a response from all players, to synchronize the start of the game.
int NetworkSynch(int numPlayers)
{	
	PrintConsoleString("Synching...");

	NetworkSendPacket(kMessageTypeSynch,nil,0,kMessagePriorityHigh,kMessageSendToAll);
	UInt32 lastSynch=0;
	UInt32 startClock;
	int start=false;
	
	int synchMessagesReceived=0;
	while(synchMessagesReceived<numPlayers)
	{
		tNetworkPacket message;
		if(NetworkReceivePacket(&message))
		{
			int cancel=false;
			if(message.what==kMessageTypeSynch)
			{
				synchMessagesReceived++;
				PrintConsoleString("Received Synch packet from player %d",message.from);
			}
			else if(message.what==kMessageTypeEndGame)
			{
				cancel=true;
				PrintConsoleString("Received Termination Packet. Aborting");
			}
			else if(message.what==kMessageTypePlayerLeft)
			{
				cancel=true;
				NetworkQueuePacket(&message);
				PrintConsoleString("A Player Left. Aborting");
			}
			else if(message.what==kMessageTypeSynchStart)
			{
				PrintConsoleString("Received Start Signal.");

				startClock=*(UInt32*)message.data;
				start=true;
			}
			else
				PrintConsoleString("Unexpected Packet");
			
			NetworkDisposePacket(&message);
			if(cancel)
				return false;
		}
		SystemPoll(true);
	}

	
	NT_MemberInfo info;
	NT_GetSelfMemberInfo(gSessionRef,&info);
	if(info.member==0)
	{	

		PrintConsoleString("Sent Start Signal.");
	
		startClock=NT_GetSessionClock(gSessionRef)+500;
		U32ChangeToNet(startClock);
		NetworkSendPacket(kMessageTypeSynchStart,&startClock,sizeof(UInt32),kMessagePriorityHigh,kMessageSendToAll);
		U32ChangeToNative(startClock);
	}
	else
	{
		float startTime=TimeGetSeconds();
		tNetworkPacket message;
		
		while(TimeGetSeconds()<startTime+kConnectionTimeOut&&!start)
		{
			if(NetworkReceivePacket(&message))
			{
				if(message.what==kMessageTypeSynchStart)
				{
					PrintConsoleString("Received Start Signal.");

					startClock=*(UInt32*)message.data;
					U32ChangeToNative(startClock);
					start=true;
				}
				else
					PrintConsoleString("Unexpected Packet");
				NetworkDisposePacket(&message);
			}		
			SystemPoll(false);
		}
		if(!start)
			return false;
	}
		
	while(NT_GetSessionClock(gSessionRef)<startClock);
	return true;
}

